from typing import TypedDict, Literal, Any
from uuid import uuid4
from copy import deepcopy

class RelationshipData(TypedDict):
	name: str
	fromTable: str
	fromColumn: str
	toTable: str
	toColumn: str
	joinOnDateBehavior: str
	crossFilteringBehavior: str | None
	state: str

class Relationship():
	name: str
	from_table: str
	from_column: str
	to_table: str
	to_column: str
	is_both_directions: bool
	reference_data: RelationshipData

	def __init__(
		self,
		from_table: str,
		from_column: str,
		to_table: str,
		to_column: str,
		is_both_directions=True
	):
		self.name = str(uuid4())
		self.from_table = from_table
		self.from_column = from_column
		self.to_table = to_table
		self.to_column = to_column
		self.is_both_directions=is_both_directions

		self.reference_data: RelationshipData = {
			"name": self.name,
			"fromTable": self.from_table,
			"fromColumn": self.from_column,
			"toTable": self.to_table,
			"toColumn": self.to_column,
			"crossFilteringBehavior": "bothDirections",
			"joinOnDateBehavior": "datePartOnly",
			"state": "ready"
		}

		if not self.is_both_directions:
			self.reference_data["crossFilteringBehavior"] = "rememberToAddThisLater"
		else:
			self.reference_data["crossFilteringBehavior"] = "bothDirections"

	def load(self, data: RelationshipData):
		self.reference_data = deepcopy(data)
		self.name = self.reference_data["name"]
		self.from_table = self.reference_data["fromTable"]
		self.from_column = self.reference_data["fromColumn"]
		self.to_table = self.reference_data["toTable"]
		self.to_column = self.reference_data["toColumn"]
		if "crossFilteringBehavior" in self.reference_data:
			if self.reference_data["crossFilteringBehavior"] == "bothDirections":
				self.is_both_directions = True
			else:
				self.is_both_directions = False
		else:
			self.is_both_directions = True

	def dump(self) -> RelationshipData:
		relationship_data = deepcopy(self.reference_data)
		relationship_data["name"] = self.name
		relationship_data["fromTable"] = self.from_table
		relationship_data["fromColumn"] = self.from_column
		relationship_data["toTable"] = self.to_table
		relationship_data["toColumn"] = self.to_column

		if not self.is_both_directions:
			relationship_data["crossFilteringBehavior"] = "rememberToAddThisLater"
		else:
			relationship_data["crossFilteringBehavior"] = "bothDirections"

		return relationship_data



import json
import inspect
from typing import TypedDict, Literal, Any
from uuid import uuid4
from copy import deepcopy
from pbit.datamodelschema.typeholder import AnnotationData
from pbit.datamodelschema.relationship import RelationshipData, Relationship
from pbit.datamodelschema.table import TableData, Table

class AccessOptionsData(TypedDict):
	legacyRedirects: bool
	returnErrorValuesAsNull: bool

class QueryGroupData(TypedDict):
	folder: str
	annotations: list[AnnotationData]

class LinguisticMetadataContent(TypedDict):
	Version: str
	Language: str
	DynamicImprovement: str

class LinguisticMetadata(TypedDict):
	contentType: str
	content: LinguisticMetadataContent

class CultureData(TypedDict):
	name: str
	linguisticMetadata: LinguisticMetadata

class ModelData(TypedDict):
	culture: str
	dataAccessOptions: AccessOptionsData
	defaultPowerBIDataSourceVersion: str
	sourceQueryCulture: str
	tables: list[TableData] | None
	relationships: list[RelationshipData] | None
	cultures: list[CultureData]
	annotations: list[AnnotationData]
	queryGroups: list[QueryGroupData] | None

class DataModelSchemaData(TypedDict):
	name: str
	compatibilityLevel: int
	model: ModelData

def get_default_table() -> TableData:
	untyped_default_table_a: Any = {
		"name": "DateTableTemplate_5975b0d6-7e08-4fc8-b30a-36e3eba94689",
		"isHidden": True,
		"isPrivate": True,
		"lineageTag": "e21631ac-035c-440b-85ef-3259718d8da4",
		"showAsVariationsOnly": None,
		"columns": [
			{
				"type": "rowNumber",
				"name": "RowNumber-2662979B-1795-4F74-8F37-6A1BA8059B61",
				"dataType": "int64",
				"isHidden": True,
				"isUnique": True,
				"isKey": True,
				"isNullable": False,
				"attributeHierarchy": {
					"state": "ready"
				}
			},
			{
				"type": "calculatedTableColumn",
				"name": "Date",
				"dataType": "dateTime",
				"isNameInferred": True,
				"isDataTypeInferred": True,
				"isHidden": True,
				"sourceColumn": "[Date]",
				"formatString": "General Date",
				"lineageTag": "889a51f9-99f0-4aa3-9e63-278d77d00ee1",
				"dataCategory": "PaddedDateTableDates",
				"summarizeBy": "none",
				"attributeHierarchy": {
					"state": "ready"
				},
				"annotations": [
					{
						"name": "SummarizationSetBy",
						"value": "User"
					}
				]
			},
			{
				"type": "calculated",
				"name": "Year",
				"dataType": "int64",
				"isDataTypeInferred": True,
				"isHidden": True,
				"expression": "YEAR([Date])",
				"formatString": "0",
				"lineageTag": "94dfb75d-4ba9-4b90-b79f-ebfe9cff087b",
				"dataCategory": "Years",
				"summarizeBy": "none",
				"attributeHierarchy": {
					"state": "ready"
				},
				"annotations": [
					{
						"name": "SummarizationSetBy",
						"value": "User"
					},
					{
						"name": "TemplateId",
						"value": "Year"
					}
				]
			},
			{
				"type": "calculated",
				"name": "MonthNo",
				"dataType": "int64",
				"isDataTypeInferred": True,
				"isHidden": True,
				"expression": "MONTH([Date])",
				"formatString": "0",
				"lineageTag": "98761376-1dab-48da-b6ff-b5851336dc89",
				"dataCategory": "MonthOfYear",
				"summarizeBy": "none",
				"attributeHierarchy": {
				"state": "ready"
				},
				"annotations": [
					{
						"name": "SummarizationSetBy",
						"value": "User"
					},
					{
						"name": "TemplateId",
						"value": "MonthNumber"
					}
				]
			},
			{
				"type": "calculated",
				"name": "Month",
				"dataType": "string",
				"isDataTypeInferred": True,
				"isHidden": True,
				"expression": "FORMAT([Date], \"MMMM\")",
				"sortByColumn": "MonthNo",
				"lineageTag": "5b5f8789-885b-493c-a768-416efed357a8",
				"dataCategory": "Months",
				"summarizeBy": "none",
				"attributeHierarchy": {
					"state": "ready"
				},
				"annotations": [
					{
						"name": "SummarizationSetBy",
						"value": "User"
					},
					{
						"name": "TemplateId",
						"value": "Month"
					}
				]
			},
			{
				"type": "calculated",
				"name": "QuarterNo",
				"dataType": "int64",
				"isDataTypeInferred": True,
				"isHidden": True,
				"expression": "INT(([MonthNo] + 2) / 3)",
				"formatString": "0",
				"lineageTag": "30244721-a2b3-4014-81b5-bf95d8f344da",
				"dataCategory": "QuarterOfYear",
				"summarizeBy": "none",
				"attributeHierarchy": {
					"state": "ready"
				},
				"annotations": [
					{
						"name": "SummarizationSetBy",
						"value": "User"
					},
					{
						"name": "TemplateId",
						"value": "QuarterNumber"
					}
				]
			},
			{
				"type": "calculated",
				"name": "Quarter",
				"dataType": "string",
				"isDataTypeInferred": True,
				"isHidden": True,
				"expression": "\"Qtr \" & [QuarterNo]",
				"sortByColumn": "QuarterNo",
				"lineageTag": "5de8ea71-ebc4-4fff-9b6d-3f3e00bf3d20",
				"dataCategory": "Quarters",
				"summarizeBy": "none",
				"attributeHierarchy": {
					"state": "ready"
				},
				"annotations": [
					{
						"name": "SummarizationSetBy",
						"value": "User"
					},
					{
						"name": "TemplateId",
						"value": "Quarter"
					}
				]
			},
			{
				"type": "calculated",
				"name": "Day",
				"dataType": "int64",
				"isDataTypeInferred": True,
				"isHidden": True,
				"expression": "DAY([Date])",
				"formatString": "0",
				"lineageTag": "9070e261-708c-4cb3-96cd-826a7cda38e8",
				"dataCategory": "DayOfMonth",
				"summarizeBy": "none",
				"attributeHierarchy": {
					"state": "ready"
				},
				"annotations": [
					{
						"name": "SummarizationSetBy",
						"value": "User"
					},
					{
						"name": "TemplateId",
						"value": "Day"
					}
				]
			}
		],
		"partitions": [
			{
				"name": "DateTableTemplate_5975b0d6-7e08-4fc8-b30a-36e3eba94689-1d04c13d-f7a1-4dae-aa9a-1a3a3400b609",
				"mode": "import",
				"state": "ready",
				"source": {
					"type": "calculated",
					"expression": "Calendar(Date(2015,1,1), Date(2015,1,1))"
				}
			}
		],
		"hierarchies": [
			{
				"name": "Date Hierarchy",
				"lineageTag": "bbe0ed27-e8ef-4e76-8a14-85f0fcb1d594",
				"state": "ready",
				"levels": [
					{
						"name": "Year",
						"ordinal": 0,
						"column": "Year",
						"lineageTag": "08329d53-fa1d-4986-8a37-f3873dee4e07"
					},
					{
						"name": "Quarter",
						"ordinal": 1,
						"column": "Quarter",
						"lineageTag": "e5d11cdc-593c-43de-996b-b6040b2a678a"
					},
					{
						"name": "Month",
						"ordinal": 2,
						"column": "Month",
						"lineageTag": "9e3fbfab-5920-462c-a4af-236b9952ac84"
					},
					{
						"name": "Day",
						"ordinal": 3,
						"column": "Day",
						"lineageTag": "1b872e0f-440e-4e03-a2b4-4e87841b1489"
					}
				],
				"annotations": [
					{
						"name": "TemplateId",
						"value": "DateHierarchy"
					}
				]
			}
		],
		"annotations": [
			{
				"name": "__PBI_TemplateDateTable",
				"value": "True"
			},
			{
				"name": "DefaultItem",
				"value": "DateHierarchy"
			}
		]
	}
	return untyped_default_table_a

# write to DataModelSchema
def remove_none_values(dct):
	# Create a list of keys to be removed (to avoid changing dictionary size during iteration)
	keys_to_remove = []
	for key_val in dct:
		value = key_val
		key = key_val
		if type(dct) != list and type(key_val) == str:
			value = dct[key_val]

		if value is None and type(dct) != list:
			keys_to_remove.append(key)
		elif inspect.isclass(value) or type(value) == list or type(value) == dict:
			remove_none_values(value)
	# Remove keys with None values
	for key in keys_to_remove:
		dct.pop(key)
	return dct


class DataModelSchema():
	relationships: list[Relationship]
	tables: list[Table]
	query_groups: list[str]
	reference_data: DataModelSchemaData
	def __init__(
		self
	):
		self.relationships = []
		self.tables = []
		self.query_groups = []	
		self.id = str(uuid4())
		self.reference_data = {
			"name": self.id,
			"compatibilityLevel": 1550,
			"model": {
				"culture": "en-US",
				"dataAccessOptions": {
					"legacyRedirects": True,
					"returnErrorValuesAsNull": True
				},
				"defaultPowerBIDataSourceVersion": "powerBI_V3",
				"sourceQueryCulture": "en-US",
				"tables": [],
				"relationships": [],
				"cultures": [
					{
						"name": "en-US",
						"linguisticMetadata": {
							"content": {
								"Version": "1.0.0",
								"Language": "en-US",
								"DynamicImprovement": "HighConfidence"
							},
							"contentType": "json"
						},
					},
				],
				"queryGroups": None,
				"annotations": [
					{
						"name": "PBI_QueryOrder",
						"value": "[\"31_2023-3-29-04c00c00p000000\"]"
					},
					{
						"name": "__PBI_TimeIntelligenceEnabled",
						"value": "1"
					},
					{
						"name": "PBIDesktopVersion",
						"value": "2.115.842.0"
					}
				]
			},
		}

		self.reference_data["model"]["tables"].append(get_default_table())

	def clear_tables(self):
		self.tables = []
		if "model" in self.reference_data:
			model_data = self.reference_data["model"]
			if "tables" in model_data:
				model_data["tables"] = []

	def clear_relationships(self): 
		self.relationships = []
		if "model" in self.reference_data:

			model_data = self.reference_data["model"]
			if "relationships" in model_data:
				model_data["relationships"] = []

	def clear_query_groups(self):
		self.query_groups = []
		if "model" in self.reference_data:
			model_data = self.reference_data["model"]
			if "queryGroups" in model_data:
				model_data["queryGroups"] = []

	def reset_model(self):
		self.clear_tables()
		self.clear_relationships()
		self.clear_query_groups()			

	def get_table_by_name(self, name: str) -> Table:
		final_table: Table | None
		for table in self.tables:
			if table.name == name:
				final_table = table
				break
		assert final_table, f"table {name} does not exist"
		return final_table
	
	def insert_query_group(self, group_name: str):
		self.query_groups.append(group_name)

	def new_relationship(
			self,
			from_table: str,
			from_column: str,
			to_table: str,
			to_column=""
		) -> Relationship:

		if to_column == "":
			to_column = from_column

		is_safe = True
		for relationship in self.relationships:
			if relationship.from_table == from_table and relationship.to_table == to_table:
				is_safe = False
			elif relationship.to_table == from_table and relationship.from_table == to_table:
				is_safe = False

		assert is_safe == True, f"there's already an active relationship between table {from_table} and {to_table}"
			
		relationship = Relationship(from_table, from_column, to_table, to_column)
		self.relationships.append(relationship)

		return relationship

	def new_table(self, name: str) -> Table:
		table = Table(name)
		self.tables.append(table)
		return table

	def load(self, schema_data: DataModelSchemaData):
		self.reference_data = deepcopy(schema_data)
		
		# print(json.dumps(self.reference_data, indent=4))

		ref_model_data: ModelData | None = None

		untyped_ref: Any = self.reference_data
		untyped_ref_data = untyped_ref["model"]
		dict_mod_data: ModelData = untyped_ref_data
		ref_model_data = dict_mod_data

		assert ref_model_data

		self.relationships = []
		self.tables = []
		self.query_groups = []

		if "relationships" in ref_model_data:
			rel_list = ref_model_data["relationships"]
			assert rel_list
			for relationship_data in rel_list:
				relationship = self.new_relationship("", "", "")
				relationship.load(relationship_data)
			ref_model_data["relationships"] = []

		if "tables" in ref_model_data:
			tab_list = ref_model_data["tables"]
			assert tab_list
			for i , table_data in enumerate(tab_list):
				table = self.new_table("")
				table.load(table_data)
			# ref_model_data["tables"] = [get_default_table()]

		if "queryGroups" in ref_model_data and ref_model_data["queryGroups"] != None:
			ref_groups = ref_model_data["queryGroups"]
			assert ref_groups
			for query_group_data in ref_groups:
				self.query_groups.append(query_group_data["folder"])
			ref_model_data["queryGroups"] = None

	def dump(self) -> DataModelSchemaData:
		data_model_schema = deepcopy(self.reference_data)

		model_data: ModelData | None = None
	
		untyped_ref: Any = data_model_schema
		untyped_ref["name"] = self.id
		untyped_ref_data = untyped_ref["model"]
		dict_mod_data: ModelData = untyped_ref_data
		model_data = dict_mod_data

		assert model_data

		if not "dataAccessOptions" in model_data:
			model_data["dataAccessOptions"] = {}

		model_data["dataAccessOptions"]["legacyRedirects"] = True
		model_data["dataAccessOptions"]["returnErrorValuesAsNull"] = True
		model_data["defaultPowerBIDataSourceVersion"] = "powerBI_V3"
		model_data["sourceQueryCulture"] = "en-US"
		if len(self.query_groups) > 0:
			query_group_data_list = []
			for i, group_name in enumerate(self.query_groups):
				query_group_data: QueryGroupData = {
					"folder": group_name,
					"annotations": [
						{
							"name": "PBI_QueryGroupOrder",
							"value": str(i)
						}
					]
				}
				query_group_data_list.append(query_group_data)

			model_data["queryGroups"] = query_group_data_list

		table_list = []

		for table in self.tables:
			table_list.append(table.dump())

		if len(table_list) > 0 or "tables" in model_data:
			model_data["tables"] = table_list

		relationship_list = []

		for relationship in self.relationships:
			relationship_list.append(relationship.dump())

		if len(relationship_list) > 0 or "relationships" in model_data:
			model_data["relationships"] = relationship_list

		return remove_none_values(data_model_schema)

def write(schema_file_path: str, data: DataModelSchemaData):
	file = open(schema_file_path, "w", encoding="utf-16-le")
	file.write(json.dumps(data, indent=4))
	file.close()

def read(schema_file_path: str) -> DataModelSchema:
	return json.loads(open(schema_file_path, "r", encoding="utf-16-le").read())




from typing import TypedDict, Literal, Any
from uuid import uuid4
from copy import deepcopy
from pandas import DataFrame
from pbit.datamodelschema.column import DaxType

class MeasureProperty(TypedDict):
	property: str

class MeasureData(TypedDict):
	name: str
	expression: str
	formatString: str
	lineageTag: str
	dataType: DaxType

class Measure():
	name: str
	id: str
	expression: str
	data_type: DaxType
	format_string: str
	reference_data: MeasureData

	def __init__(
		self,
		name: str,
		data_type: DaxType
	):
		self.name = name
		self.id = str(uuid4())
		self.expression = ""
		self.data_type = data_type
		self.format_string = ""
		self.reference_data: MeasureData = {
			"name": self.name,
			"expression": self.expression,
			"formatString": self.format_string,
			"lineageTag": self.id,
			"dataType": self.data_type,
		}

	def set_format(self, format: str):
		self.format_string = format

	def set_expression(self, expression: str, data_type: DaxType):
		self.expression = expression
		self.data_type = data_type

	def set_to_retention_rate_tracker(self, user_table_name: str, is_retained_column_name: str):
		expression = f"COUNTROWS(FILTER({user_table_name}, {user_table_name}[{is_retained_column_name}]=True))/COUNTROWS({user_table_name})"
		self.set_expression(expression, "double")
		self.set_format("0.00%;-0.00%;0.00%")

	def load(self, data: MeasureData):
		self.reference_data = deepcopy(data)
		self.name = self.reference_data["name"]
		self.expression = self.reference_data["expression"]
		self.format_string = self.reference_data["formatString"]
		self.id = self.reference_data["lineageTag"]
		self.data_type = self.reference_data["dataType"]

	def dump(self) -> MeasureData:
		measure_data = deepcopy(self.reference_data)
		measure_data["name"] = self.name
		measure_data["expression"] = self.expression
		measure_data["formatString"] = self.format_string
		measure_data["lineageTag"] = self.id
		measure_data["dataType"] = self.data_type

		return measure_data



from typing import TypedDict, Literal, Any
from uuid import uuid4
from copy import deepcopy
from pbit.datamodelschema.column import DaxType
from pbit.datamodelschema.powerquery import PowerQuery, MType, from_dax_type_to_m_type

class PartitionSourceData(TypedDict):
	type: str
	expression: list[str] | str

class PartitionData(TypedDict):
	name: str | None
	mode: str
	state: str
	queryGroup: str | None
	source: PartitionSourceData

class Partition():
	name: str | None
	language: str
	power_query: PowerQuery
	query_group: str | None
	reference_data: PartitionData

	def __init__(
		self,
		name: str | None = None,
		language: str = "m",
		query_group: str | None = None
	):
		self.name = name
		self.language = language
		self.power_query = PowerQuery()
		self.query_group = query_group
		self.reference_data: PartitionData = {
			"name": self.name,
			"mode": "import",
			"state": "ready",
			"queryGroup": self.query_group,
			"source": {
				"type": self.language,
				"expression": []
			}
		}

	def set_to_json_reader(
		self,
		relative_json_path: str,
		dax_types: dict[str, DaxType]
	):
		self.power_query.insert_read_file_cmd(relative_json_path)
		self.power_query.insert_load_as_json_cmd()
		self.power_query.insert_table_from_list_cmd()
		self.power_query.insert_expand_from_record_cmd(
			target_column="Column1",
			conversion_table=list(from_dax_type_to_m_type(dax_types).keys())
		)
		m_dax_types = from_dax_type_to_m_type(dax_types)
		self.power_query.insert_transform_dax_types_cmd(m_dax_types)

	def load(self, data: PartitionData):
		self.reference_data = deepcopy(data)
		if "name" in self.reference_data:
			self.name = self.reference_data["name"]
		if "queryGroup" in self.reference_data:
			self.query_group = self.reference_data["queryGroup"]
		else:
			self.query_group = None
		self.language =self.reference_data["source"]["type"]
		
	def dump(self) -> PartitionData:
		partition_data = deepcopy(self.reference_data)
		partition_data["name"] = self.name
		partition_data["queryGroup"] = self.query_group
		partition_data["source"]["type"] = self.language
		if len(self.power_query.commands) > 0:
			partition_data["source"]["expression"] = self.power_query.dump()

		return partition_data


from typing import TypedDict, Literal, Any
from uuid import uuid4
from copy import deepcopy
from pbit.datamodelschema.dax import DaxType
from pbit.datamodelschema.typeholder import AnnotationData

class ColumnAttributeHierarchyData(TypedDict):
	state: str

class ColumnData(TypedDict):
	type: DaxType | None
	lineageTag: str | None
	summarizeBy: str | None
	sourceColumn: str | None
	name: str
	dataType: str
	formatString: str | None
	dataCategory: str | None
	isNameInferred: bool | None
	expression: str | list[str] | None
	isDataTypeInferred: bool | None
	isHidden: bool | None
	isUnique: bool | None
	sortByColumn: str | None
	isKey: bool | None
	isNullable: bool | None
	attributeHierarchy: ColumnAttributeHierarchyData
	annotations: list[AnnotationData] | None

class Column():
	name: str
	id: str
	data_type: str
	source_column: str | None
	reference_data: ColumnData

	def __init__(
		self,
		name: str,
		dataType: str,
		source_column: str | None = None
	):
		self.name = name
		self.id = str(uuid4())
		self.data_type = dataType
		self.source_column = source_column
		self.reference_data: ColumnData = {
			"formatString": None,
			"dataCategory": None,
			"isNameInferred": None,
			"expression": None,
			"isDataTypeInferred": None,
			"sortByColumn": None,
			"type": None,
			"isHidden": None,
			"isUnique": None,
			"isKey": None,
			"isNullable": None,
			"name": self.name,
			"dataType": self.data_type,
			"sourceColumn": self.source_column,
			"lineageTag": self.id,
			"summarizeBy": "none",
			"attributeHierarchy": {
				"state": "ready"
			},
			"annotations": [
				{
					"name": "SummarizationSetBy",
					"value": "Automatic"
				}
			]
		}

	def set_as_bin(self, target_table_name: str, target_column_name: str, increment: float, bin_name: str | None = None, data_type: DaxType="double"):
		final_name: str = ""
		if bin_name:
			final_name = bin_name
		else:
			final_name = target_column_name + "_bin"
		
		reference_data: Any = {
			"type": "calculated",
			"name": final_name,
			"dataType": data_type,
			"isDataTypeInferred": True,
			"expression": [
				"IF(",
				f"\tISBLANK('{target_table_name}'[{target_column_name}]),",
				"\tBLANK(),",
				"\tIF(",
				f"\t\t'{target_table_name}'[{target_column_name}] >= 0,",
				f"\t\tROUNDDOWN('{target_table_name}'[{target_column_name}] / {increment}, 0) * {increment},",
				f"\t\tROUNDUP('{target_table_name}'[{target_column_name}] / {increment}, 0) * {increment}",
				"\t)",
				")"
			],
			"lineageTag": self.id,
			"summarizeBy": "none",
			"attributeHierarchy": {
				"state": "ready"
			},
			"extendedProperties": [
				{
					"type": "json",
					"name": "GroupingMetadata",
					"value": {
						"version": 0,
						"groupedColumns": [
							{
								"Column": {
									"Expression": {
										"SourceRef": {
											"Entity": target_table_name
										}
									},
									"Property": target_column_name
								}
							}
						],
						"binningMetadata": {
							"binSize": {
								"value": 5.0,
								"unit": 0
							}
						}
					}
				}
			],
			"annotations": [
				{
					"name": "GroupingDesignState",
					"value": "{\"Version\":0,\"Sources\":[{\"Name\":\"p\",\"Entity\":\""
						+target_table_name+"\"}],\"GroupedColumns\":[{\"Column\":{\"Expression\":{\"SourceRef\":{\"Source\":\"p\"}},\"Property\":\""
						+target_column_name+"\"}}],\"BinItem\":{\"Expression\":{\"Floor\":{\"Expression\":{\"Column\":{\"Expression\":{\"SourceRef\":{\"Source\":\"p\"}},\"Property\":\""
						+target_column_name+"\"}},\"Size\":"
						+str(increment)+"}}}}"
				},
				{
					"name": "SummarizationSetBy",
					"value": "Automatic"
				},
				{
					"name": "PBI_FormatHint",
					"value": "{\"isGeneralNumber\":true}"
				}
			]
		}
		self.load(reference_data)

	def load(self, data: ColumnData):
		self.reference_data = deepcopy(data)
		self.name = self.reference_data["name"]
		self.data_type = self.reference_data["dataType"]
		if "sourceColumn" in self.reference_data:
			source_column_ref = self.reference_data["sourceColumn"]
			assert source_column_ref
			self.source_column = source_column_ref
		
		if "lineageTag" in self.reference_data:
			lin_tag = self.reference_data["lineageTag"]
			assert lin_tag
			self.id = lin_tag

	def dump(self) -> ColumnData:
		column_data: ColumnData = deepcopy(self.reference_data)
		column_data["name"] = self.name
		column_data["dataType"] = self.data_type
		if column_data["dataType"] == "any":
			column_data["dataType"] = "string"

		column_data["sourceColumn"] = self.source_column
		column_data["lineageTag"] = self.id

		return column_data



from typing import TypedDict, Literal, Any, Callable
from uuid import uuid4
from pbit.datamodelschema.column import DaxType
import os

PRIOR_PREFIX: str = "~!PRIOR_TABLE_KEY!~"

MType = Literal[
	"Int64.Type",
	"Currency.Type",
	"Percentage.Type",
	"type datetime", 
	"type date",
	"type time",
	"type datetimezone",
	"type duration",
	"type logical",
	"type binary",
	"type number",
	"type any",
	"type text"
]

M_TYPE_TO_COLUMN_TYPE: dict[MType, DaxType] = {
	"type text": "string",
	"type logical": "boolean",
	"type datetime": "dateTime",
	"type any": "string",
	"Int64.Type": "int64",
	"type number": "double",
}

def from_m_type_to_dax_type(m_type_dict: dict[str, MType]) -> dict[str, DaxType]:
	out: dict[str, DaxType] = {}
	for key in m_type_dict:
		out[key] = M_TYPE_TO_COLUMN_TYPE[m_type_dict[key]]
	return out

def from_dax_type_to_m_type(dax_type_dict: dict[str, DaxType]) -> dict[str, MType]:
	out: dict[str, MType] = {}
	for key in dax_type_dict:

		if dax_type_dict[key] == "any":
			out[key] = "type any"
		else:
			for m_type, col_type in M_TYPE_TO_COLUMN_TYPE.items():
				if dax_type_dict[key] == col_type:
					out[key] = m_type
	return out

def list_to_m_str(list: list, wrap_values_as_str: bool) -> str:
	out = "{"
	for i, val in enumerate(list):
		if wrap_values_as_str:
			out += "\""+str(val)+"\""
		else:
			out += str(val)
		if i < len(list)-1:
			out += ", "
	out += "}"
	return out

def dict_to_m_str(dict: dict, wrap_keys_as_str: bool, wrap_values_as_str: bool) -> str:
	out = "{"
	key_index = 0
	key_length = len(list(dict.keys()))
	for k, v in dict.items():
		key_index += 1
		key_entry = k
		val_entry = v
		if wrap_keys_as_str:
			key_entry = "\"" + k + "\""

		if wrap_values_as_str:
			val_entry = "\"" + v + "\""

		out += list_to_m_str([key_entry, val_entry], False)
		if key_index < key_length:
			out += ","

	out += "}"
	return out

class Command():
	id: str
	name: str
	function_text: str
	parameter_order: list[str]
	parameters: dict[str, str | int | None]
	get_command_name: Callable[[str], str]
	def __init__(
		self,
		function_text: str,
		parameter_order: list[str],
		parameters: dict[str, str | int | None],
		name: str | None = None
	):
		self.id = str(uuid4())
		self.function_text = function_text
		if name == None:
			self.name = self.function_text.replace(".", " ")
		else:
			assert name
			self.name = name
		self.parameter_order = parameter_order
		self.parameters = parameters

	def dump(self, command_name: str | None = None, include_comma: bool = True) -> str:
		command_str = self.function_text + "("
		for i, key in enumerate(self.parameter_order):
			command_str += " "
			if key in self.parameters and self.parameters[key] != None:
				val: str | int | None = self.parameters[key]
				assert val

				command_str += str(val)

			else:
				command_str += "null"

			if i < len(self.parameter_order)-1:
				command_str += ","

		command_str += ")"

		if include_comma:
			command_str += ","

		if command_name != None:
			return f"#\"{command_name}\" = " + command_str
		else:
			return command_str

class PowerQuery():
	commands: list[Command]
	def __init__(self):
		self.commands = []

	def dump(self) -> list[str]:
		bases: dict[str, int] = {}
		query = "let"
		final_command_name = ""

		for i, command in enumerate(self.commands):
			command_name = command.name
			if not command_name in bases:
				bases[command_name] = 1
				command_name += "1"
			else:
				bases[command_name] += 1
				command_name += str(bases[command_name]+1)
			command_str = command.dump(command_name, i < len(self.commands)-1)
			command_str = command_str.replace(PRIOR_PREFIX, final_command_name)
			final_command_name = command_name
			query += "\n\t"+command_str
			
		query += "\nin"
		query += f"\n\t#\"{final_command_name}\""

		return query.split("\n")

	def insert_cmd(
		self,
		function_text: str,
		parameter_order: list[str],
		parameters: dict[str, str | int | None],
		name: str | None = None) -> Command:
		command = Command(function_text, parameter_order, parameters, name)
		self.commands.append(command)
		return command


	def insert_read_file_cmd(
		self,
		relative_path: str,
		name: str | None = None
	) -> Command:
		command = self.insert_cmd(
			function_text = "File.Contents",
			parameter_order = ["path", "options"],
			parameters = {
				"path": ("\"" + os.path.abspath(relative_path) + "\"").replace("\\", "/")
			},
			name = name
		)
		return command

	def insert_load_as_json_cmd(
		self,
		source = "#\""+PRIOR_PREFIX+"\"",
		encoding: int | None = None,
		name: str | None = None
	) -> Command:
		command = self.insert_cmd(
			function_text = "Json.Document",
			parameter_order = ["source", "encoding"],
			parameters = {
				"source": source,
				"encoding": encoding,
			},
			name = name
		)
		return command
	def insert_load_as_csv_cmd(
		self,
		source = "#\""+PRIOR_PREFIX+"\"",
		column_names: list | None = None,
		delimiter: str | None = None,
		extra_values: str | None = "ExtraValues.Error",
		encoding: int | None = None,
		name: str | None = None
	) -> Command:

		column_list_string: str | None = None
		if column_names and len(column_names) > 0:
			assert column_names
			column_list_string = list_to_m_str(column_names, True)

		command = self.insert_cmd(
			function_text = "Csv.Document",
			parameter_order = [
				"source", 
				"columns", 
				"delimiter", 
				"extra_values", 
				"encoding"
			],
			parameters = {
				"source": source,
				"columns": column_list_string,
				"delimiter": delimiter,
				"extra_values": extra_values,
				"encoding": encoding,
			},
			name = name
		)
		return command

	def insert_table_from_list_cmd(
		self,
		source = "#\""+PRIOR_PREFIX+"\"",
		splitter = "Splitter.SplitByNothing()",
		columns: str | None = None,
		default: str | None = None,
		extra_values: str | None = "ExtraValues.Error",
		name: str | None = None
	) -> Command:
		command = self.insert_cmd(
			function_text = "Table.FromList",
			parameter_order = [
				"source", 
				"splitter", 
				"columns", 
				"default", 
				"extra_values"
			],
			parameters = {
				"source": source,
				"splitter": splitter,
				"columns": columns,
				"default": default,
				"extra_values": extra_values,
			},
			name = name
		)
		return command

	def insert_expand_from_record_cmd(
		self,
		target_column: str,
		conversion_table: dict[str, str] | list[str],
		source = "#\""+PRIOR_PREFIX+"\"",
		name: str | None = None
	) -> Command:
		field_names_str: str = "{}"
		column_names_str: str | None = None
		if type(conversion_table) == dict:
			field_names = list(conversion_table.keys())
			new_column_names = []
			
			for field_name in field_names:
				new_column_names.append(conversion_table[field_name])
			
			field_names_str = list_to_m_str(field_names, True)
			column_names_str = list_to_m_str(new_column_names, True)

		else:
			assert type(conversion_table) == list
			field_names_str = list_to_m_str(conversion_table, True)

		command = self.insert_cmd(
			function_text = "Table.ExpandRecordColumn",
			parameter_order = [
				"source", 
				"column",
				"field_names",
				"new_column_names"
			],
			parameters = {
				"source": source,
				"column": "\""+target_column+"\"",
				"field_names": field_names_str,
				"new_column_names": column_names_str
			},
			name = name
		)
		return command	

	def insert_transform_dax_types_cmd(
		self,
		transformations: dict[str, MType],
		source = "#\""+PRIOR_PREFIX+"\"",
		culture: str | None = None,
		name: str | None = None
	) -> Command:
		command = self.insert_cmd(
			function_text = "Table.TransformColumnTypes",
			parameter_order = [
				"source", 
				"type_transformations",
				"culture"
			],
			parameters = {
				"source": source,
				"type_transformations": dict_to_m_str(transformations, True, False),
				"culture": culture
			},
			name = name
		)
		return command	


from typing import TypedDict, Literal, Any
from uuid import uuid4
from copy import deepcopy
import pandas as pd
from pbit.datamodelschema.dax import DaxType
from pbit.datamodelschema.column import Column, ColumnData 
from pbit.datamodelschema.partition import Partition, PartitionData
from pbit.datamodelschema.measure import Measure, MeasureData
from pbit.datamodelschema.typeholder import AnnotationData
from pbit.datamodelschema.powerquery import MType

class HierarchyLevelData(TypedDict):
	name: str
	ordinal: int
	column: str
	linageTag: str

class HierarchyData(TypedDict):
	name: str
	lineageTag: str
	state: str
	levels: list[HierarchyLevelData]

class TableData(TypedDict):
	name: str | None
	isHidden: bool | None
	isPrivate: bool | None
	showAsVariationsOnly: bool | None
	lineageTag: str
	columns: list[ColumnData]
	partitions: list[PartitionData]
	hierarchies: list[HierarchyData] | None
	annotations: list[AnnotationData]
	measures: list[MeasureData] | None

class Table():
	name: str
	id: str
	columns: list[Column]
	partitions: list[Partition]
	measures: list[Measure]
	reference_data: TableData

	def __init__(
		self,
		name: str
	):
		self.name = name
		self.id = str(uuid4())
		self.columns: list[Column] = []
		self.partitions: list[Partition] = []
		self.measures: list[Measure] = []
		ref_data: Any = {
			"name": self.name,
			"lineageTag": self.id,
			"isHidden": None,
			"isPrivate": None,
			"showAsVariationsOnly": None,
			"hierarchies": None,
			"columns": [
				{
					"formatString": None,
					"dataCategory": None,
					"isNameInferred": None,
					"expression": None,
					"isDataTypeInferred": None,
					"sortByColumn": None,
					"type": "rowNumber",
					"name": "RowNumber-2662979B-1795-4F74-8F37-6A1BA8059B61",
					"dataType": "int64",
					"isHidden": True,
					"isUnique": True,
					"isKey": True,
					"isNullable": False,
					"lineageTag": None,
					"summarizeBy": None,
					"sourceColumn": None,
					"annotations": [
						{
							"name": "PBI_ResultType",
							"value": "Table"
						}
					],
					"attributeHierarchy": {
						"state": "ready"
					},
				},
			],
			"partitions": [],
			"measures": None,
			"annotations": [
				{
					"name": "PBI_ResultType",
					"value": "Table"
				}
			]
		}
		self.reference_data: TableData = ref_data

	def bind_to_json(
		self, 
		relative_json_path: str, 
		type_dictionary: dict[str, DaxType], 
		group_name: str | None = None
	):
		self.new_partition(group_name).set_to_json_reader(relative_json_path, type_dictionary)
		
		# add columns
		for name, col_type in type_dictionary.items():
			self.new_column(name, col_type, name)

	def new_partition(self, name="", language="m", query_group: str | None = None) -> Partition:
		if name == "":
			name = self.name

		partition = Partition(name, language, query_group)
		self.partitions.append(partition)
		return partition

	def get_column_by_name(self, name: str) -> Column:
		final_column: Column | None = None
		# print("column count: ", len(self.columns))
		for column in self.columns:
			# print(name, " =? ", column.name)
			if column.name == name:
				final_column = column
				break

		assert final_column, f"column {name} does not exist in table {self.name}"
		return final_column

	def new_column(
		self,
		name: str,
		data_type: DaxType,
		source_column: None | str = None
	) -> Column:
		if source_column == None and name != "":
			source_column = name
		column = Column(name, data_type, source_column)
		self.columns.append(column)
		return column

	def new_bin(self, target_column_name: str, increment: float, target_table_name: str | None = None,  bin_name: str | None = None, data_type: DaxType ="double"):
		final_table_name = ""
		if target_table_name:
			final_table_name = target_table_name
		else:
			final_table_name = self.name

		column = self.new_column("", data_type)
		column.set_as_bin(final_table_name, target_column_name, increment, bin_name, data_type)

	def new_measure(self, name: str) -> Measure:
		measure = Measure(name, "any")
		self.measures.append(measure)
		return measure

	def load(self, data: TableData):
		self.reference_data = deepcopy(data)
		if "name" in self.reference_data:
			name = self.reference_data["name"]
			assert name
			self.name = name

		self.id = self.reference_data["lineageTag"]

		for column_data in self.reference_data["columns"]:
			column = self.new_column("", "any")
			column.load(column_data)
		self.reference_data["columns"] = []

		for partition_data in self.reference_data["partitions"]:
			partition = self.new_partition("")
			partition.load(partition_data)

		self.reference_data["partitions"] = []

	def dump(self) -> TableData:
		table_data: TableData = deepcopy(self.reference_data)

		for column in self.columns:
			table_data["columns"].append(column.dump())

		for partition in self.partitions:
			table_data["partitions"].append(partition.dump())

		if len(self.measures) > 0:
			measure_list = []
			for measure in self.measures:
				measure_list.append(measure.dump())
			table_data["measures"] = measure_list

		return table_data



import shutil
import os
import zipfile
from tempfile import TemporaryDirectory
from .datamodelschema import read as read_pbit
from .datamodelschema import write as write_pbit
from .datamodelschema import DataModelSchema, DataModelSchemaData

def pack_pbti_file(dir_path: str, out_path: str):
	zip_path = dir_path + ".zip"

	# zip directory back up
	if os.path.exists(zip_path):
		os.remove(zip_path)
	zip_out = zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED)
	for root, _, files in os.walk(dir_path):
		for file in files:
			# Construct the full file path
			file_path = os.path.join(root, file)
			# Construct the relative file path (relative to the directory being zipped)
			relative_path = os.path.relpath(file_path, dir_path)
			# Add the file to the ZIP file with the relative path as the archive name
			zip_out.write(file_path, arcname=relative_path)
	zip_out.close()

	# convert to pbit file in final position
	if os.path.exists(out_path):
		os.remove(out_path)
	shutil.move(zip_path, out_path)

def unpack_pbti_file(file_path: str, out_path: str):
	base = os.path.split(out_path)[0]
	zip_path = base + ".zip"

	# convert to zip file
	if os.path.exists(zip_path):
		os.remove(zip_path)
	shutil.copy(file_path, zip_path)

	# unzip and extract data
	if os.path.exists(out_path):
		shutil.rmtree(out_path)
	zip_ref = zipfile.ZipFile(zip_path, 'r')
	zip_ref.extractall(out_path)
	zip_ref.close()
		
def write_model(pbit_file_path: str, source: DataModelSchemaData | DataModelSchema):
	with TemporaryDirectory() as temp_dir_path:
		pbti_dir = temp_dir_path + "/pbit"
		unpack_pbti_file(pbit_file_path, pbti_dir)
		if type(source) == DataModelSchema:
			write_pbit(pbti_dir+"/DataModelSchema", source.dump())
		elif type(source) == DataModelSchemaData:
			write_pbit(pbti_dir+"/DataModelSchema", source)
		pack_pbti_file(pbti_dir, pbit_file_path)	

def read_model(pbit_file_path: str) -> DataModelSchemaData:
	with TemporaryDirectory() as temp_dir_path:
		pbti_dir = temp_dir_path + "/pbit"
		unpack_pbti_file(pbit_file_path, pbti_dir)
		return read_pbit(pbti_dir+"/DataModelSchema")

def load_model(pbit_file_path: str) -> DataModelSchema:
	return DataModelSchema(read_model(pbit_file_path))
